@@ -43,19 +43,21 @@
 
 bool sna_composite_create(struct sna *sna)
 {
-	xRenderColor color ={ 0 };
+	xRenderColor color = { 0 };
 	int error;
 
-	if (!can_render(sna))
-		return true;
-
 	sna->clear = CreateSolidPicture(0, &color, &error);
 	return sna->clear != NULL;
 }
 
 void sna_composite_close(struct sna *sna)
 {
-	FreePicture(sna->clear, 0);
+	DBG(("%s\n", __FUNCTION__));
+
+	if (sna->clear) {
+		FreePicture(sna->clear, 0);
+		sna->clear = NULL;
+	}
 }
 
 static inline bool
@@ -214,10 +216,13 @@ sna_compute_composite_region(RegionPtr region,
 			pixman_region_fini (region);
 			return FALSE;
 		}
-		DBG(("%s: clip against src: (%d, %d), (%d, %d)\n",
-		     __FUNCTION__,
-		     region->extents.x1, region->extents.y1,
-		     region->extents.x2, region->extents.y2));
+		DBG(("%s: clip against src (%dx%d clip=%d): (%d, %d), (%d, %d)\n",
+		       __FUNCTION__,
+		       src->pDrawable ? src->pDrawable->width : 0,
+		       src->pDrawable ? src->pDrawable->height : 0,
+		       src->clientClipType,
+		       region->extents.x1, region->extents.y1,
+		       region->extents.x2, region->extents.y2));
 
 		if (src->alphaMap) {
 			if (!clip_to_src(region, src->alphaMap,
@@ -400,7 +405,7 @@ static void _assert_pixmap_contains_box(PixmapPtr pixmap, BoxPtr box, const char
 static void apply_damage(struct sna_composite_op *op, RegionPtr region)
 {
 	DBG(("%s: damage=%p, region=%d [(%d, %d), (%d, %d) + (%d, %d)]\n",
-	     __FUNCTION__, op->damage, REGION_NUM_RECTS(region),
+	     __FUNCTION__, op->damage, (int)RegionNumRects(region),
 	     region->extents.x1, region->extents.y1,
 	     region->extents.x2, region->extents.y2,
 	     op->dst.x, op->dst.y));
@@ -438,6 +443,164 @@ static inline bool use_cpu(PixmapPtr pixmap, struct sna_pixmap *priv,
 	return (priv->create & KGEM_CAN_CREATE_GPU) == 0;
 }
 
+static void validate_source(PicturePtr picture)
+{
+#if XORG_VERSION_CURRENT >= XORG_VERSION_NUMERIC(1,10,99,901,0)
+	miCompositeSourceValidate(picture);
+#else
+	miCompositeSourceValidate(picture,
+				  0, 0,
+				  picture->pDrawable ? picture->pDrawable->width : 0,
+				  picture->pDrawable ? picture->pDrawable->height : 0);
+#endif
+}
+
+void
+sna_composite_fb(CARD8 op,
+		 PicturePtr src,
+		 PicturePtr mask,
+		 PicturePtr dst,
+		 RegionPtr region,
+		 INT16 src_x, INT16 src_y,
+		 INT16 msk_x, INT16 msk_y,
+		 INT16 dst_x, INT16 dst_y,
+		 CARD16 width, CARD16 height)
+{
+	pixman_image_t *src_image, *mask_image, *dest_image;
+	int src_xoff, src_yoff;
+	int msk_xoff, msk_yoff;
+	int dst_xoff, dst_yoff;
+	int16_t tx, ty;
+	unsigned flags;
+
+	DBG(("%s -- op=%d, fallback dst=(%d, %d)+(%d, %d), size=(%d, %d): region=((%d,%d), (%d, %d))\n",
+	     __FUNCTION__, op,
+	     dst_x, dst_y,
+	     dst->pDrawable->x, dst->pDrawable->y,
+	     width, height,
+	     region->extents.x1, region->extents.y1,
+	     region->extents.x2, region->extents.y2));
+
+	if (src->pDrawable) {
+		DBG(("%s: fallback -- move src to cpu\n", __FUNCTION__));
+		if (!sna_drawable_move_to_cpu(src->pDrawable,
+					      MOVE_READ))
+			return;
+
+		if (src->alphaMap &&
+		    !sna_drawable_move_to_cpu(src->alphaMap->pDrawable,
+					      MOVE_READ))
+			return;
+	}
+
+	validate_source(src);
+
+	if (mask) {
+		if (mask->pDrawable) {
+			DBG(("%s: fallback -- move mask to cpu\n", __FUNCTION__));
+			if (!sna_drawable_move_to_cpu(mask->pDrawable,
+						      MOVE_READ))
+				return;
+
+			if (mask->alphaMap &&
+			    !sna_drawable_move_to_cpu(mask->alphaMap->pDrawable,
+						      MOVE_READ))
+				return;
+		}
+
+		validate_source(mask);
+	}
+
+	DBG(("%s: fallback -- move dst to cpu\n", __FUNCTION__));
+	if (op <= PictOpSrc && !dst->alphaMap)
+		flags = MOVE_WRITE | MOVE_INPLACE_HINT;
+	else
+		flags = MOVE_WRITE | MOVE_READ;
+	if (!sna_drawable_move_region_to_cpu(dst->pDrawable, region, flags))
+		return;
+	if (dst->alphaMap &&
+	    !sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, flags))
+		return;
+
+	if (mask == NULL &&
+	    src->pDrawable &&
+	    dst->pDrawable->bitsPerPixel >= 8 &&
+	    src->filter != PictFilterConvolution &&
+	    (op == PictOpSrc || (op == PictOpOver && !PICT_FORMAT_A(src->format))) &&
+	    (dst->format == src->format || dst->format == alphaless(src->format)) &&
+	    sna_transform_is_integer_translation(src->transform, &tx, &ty)) {
+		PixmapPtr dst_pixmap = get_drawable_pixmap(dst->pDrawable);
+		PixmapPtr src_pixmap = get_drawable_pixmap(src->pDrawable);
+		int16_t sx = src_x + tx - (dst->pDrawable->x + dst_x);
+		int16_t sy = src_y + ty - (dst->pDrawable->y + dst_y);
+
+		assert(src->pDrawable->bitsPerPixel == dst->pDrawable->bitsPerPixel);
+		assert(src_pixmap->drawable.bitsPerPixel == dst_pixmap->drawable.bitsPerPixel);
+
+		if (region->extents.x1 + sx >= 0 &&
+		    region->extents.y1 + sy >= 0 &&
+		    region->extents.x2 + sx <= src->pDrawable->width &&
+		    region->extents.y2 + sy <= src->pDrawable->height) {
+			BoxPtr box = RegionRects(region);
+			int nbox = RegionNumRects(region);
+
+			sx += src->pDrawable->x;
+			sy += src->pDrawable->y;
+			if (get_drawable_deltas(src->pDrawable, src_pixmap, &tx, &ty))
+				sx += tx, sy += ty;
+
+			assert(region->extents.x1 + sx >= 0);
+			assert(region->extents.x2 + sx <= src_pixmap->drawable.width);
+			assert(region->extents.y1 + sy >= 0);
+			assert(region->extents.y2 + sy <= src_pixmap->drawable.height);
+
+			get_drawable_deltas(dst->pDrawable, dst_pixmap, &tx, &ty);
+
+			assert(nbox);
+			do {
+				assert(box->x1 + sx >= 0);
+				assert(box->x2 + sx <= src_pixmap->drawable.width);
+				assert(box->y1 + sy >= 0);
+				assert(box->y2 + sy <= src_pixmap->drawable.height);
+
+				assert(box->x1 + tx >= 0);
+				assert(box->x2 + tx <= dst_pixmap->drawable.width);
+				assert(box->y1 + ty >= 0);
+				assert(box->y2 + ty <= dst_pixmap->drawable.height);
+
+				assert(box->x2 > box->x1 && box->y2 > box->y1);
+
+				memcpy_blt(src_pixmap->devPrivate.ptr,
+					   dst_pixmap->devPrivate.ptr,
+					   dst_pixmap->drawable.bitsPerPixel,
+					   src_pixmap->devKind,
+					   dst_pixmap->devKind,
+					   box->x1 + sx, box->y1 + sy,
+					   box->x1 + tx, box->y1 + ty,
+					   box->x2 - box->x1, box->y2 - box->y1);
+				box++;
+			} while (--nbox);
+
+			return;
+		}
+	}
+
+	src_image = image_from_pict(src, FALSE, &src_xoff, &src_yoff);
+	mask_image = image_from_pict(mask, FALSE, &msk_xoff, &msk_yoff);
+	dest_image = image_from_pict(dst, TRUE, &dst_xoff, &dst_yoff);
+
+	if (src_image && dest_image && !(mask && !mask_image))
+		sna_image_composite(op, src_image, mask_image, dest_image,
+				    src_x + src_xoff, src_y + src_yoff,
+				    msk_x + msk_xoff, msk_y + msk_yoff,
+				    dst_x + dst_xoff, dst_y + dst_yoff,
+				    width, height);
+
+	free_pixman_pict(src, src_image);
+	free_pixman_pict(mask, mask_image);
+	free_pixman_pict(dst, dest_image);
+}
+
 void
 sna_composite(CARD8 op,
 	      PicturePtr src,
@@ -452,7 +615,6 @@ sna_composite(CARD8 op,
 	struct sna *sna = to_sna_from_pixmap(pixmap);
 	struct sna_pixmap *priv;
 	struct sna_composite_op tmp;
-	unsigned flags;
 	RegionRec region;
 	int dx, dy;
 
@@ -471,7 +633,8 @@ sna_composite(CARD8 op,
 	if (op == PictOpClear) {
 		DBG(("%s: discarding source and mask for clear\n", __FUNCTION__));
 		mask = NULL;
-		src = sna->clear;
+		if (sna->clear)
+			src = sna->clear;
 	}
 
 	if (mask && sna_composite_mask_is_opaque(mask)) {
@@ -510,7 +673,7 @@ sna_composite(CARD8 op,
 	}
 
 	if (use_cpu(pixmap, priv, op, width, height) &&
-	    !picture_is_gpu(src) && !picture_is_gpu(mask)) {
+	    !picture_is_gpu(sna, src) && !picture_is_gpu(sna, mask)) {
 		DBG(("%s: fallback, dst pixmap=%ld is too small (or completely damaged)\n",
 		     __FUNCTION__, pixmap->drawable.serialNumber));
 		goto fallback;
@@ -527,14 +690,17 @@ sna_composite(CARD8 op,
 	     get_drawable_dx(dst->pDrawable),
 	     get_drawable_dy(dst->pDrawable)));
 
-	if (op <= PictOpSrc) {
+	if (op <= PictOpSrc && priv->cpu_damage) {
 		int16_t x, y;
 
-		get_drawable_deltas(dst->pDrawable, pixmap, &x, &y);
-		if (x|y)
+		if (get_drawable_deltas(dst->pDrawable, pixmap, &x, &y))
 			pixman_region_translate(&region, x, y);
 
 		sna_damage_subtract(&priv->cpu_damage, &region);
+		if (priv->cpu_damage == NULL) {
+			list_del(&priv->flush_list);
+			priv->cpu = false;
+		}
 
 		if (x|y)
 			pixman_region_translate(&region, -x, -y);
@@ -557,72 +723,24 @@ sna_composite(CARD8 op,
 		tmp.box(sna, &tmp, &region.extents);
 	else
 		tmp.boxes(sna, &tmp,
-			  REGION_BOXPTR(&region),
-			  REGION_NUM_RECTS(&region));
+			  RegionBoxptr(&region),
+			  RegionNumRects(&region));
 	apply_damage(&tmp, &region);
 	tmp.done(sna, &tmp);
 
 	goto out;
 
 fallback:
-	DBG(("%s -- fallback dst=(%d, %d)+(%d, %d), size=(%d, %d): region=((%d,%d), (%d, %d))\n",
-	     __FUNCTION__,
-	     dst_x, dst_y,
-	     dst->pDrawable->x, dst->pDrawable->y,
-	     width, height,
-	     region.extents.x1, region.extents.y1,
-	     region.extents.x2, region.extents.y2));
-	if (op <= PictOpSrc && !dst->alphaMap)
-		flags = MOVE_WRITE | MOVE_INPLACE_HINT;
-	else
-		flags = MOVE_WRITE | MOVE_READ;
-	DBG(("%s: fallback -- move dst to cpu\n", __FUNCTION__));
-	if (!sna_drawable_move_region_to_cpu(dst->pDrawable, &region, flags))
-		goto out;
-	if (dst->alphaMap &&
-	    !sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, flags))
-		goto out;
-	if (src->pDrawable) {
-		DBG(("%s: fallback -- move src to cpu\n", __FUNCTION__));
-		if (!sna_drawable_move_to_cpu(src->pDrawable,
-					      MOVE_READ))
-			goto out;
-
-		if (src->alphaMap &&
-		    !sna_drawable_move_to_cpu(src->alphaMap->pDrawable,
-					      MOVE_READ))
-			goto out;
-	}
-	if (mask && mask->pDrawable) {
-		DBG(("%s: fallback -- move mask to cpu\n", __FUNCTION__));
-		if (!sna_drawable_move_to_cpu(mask->pDrawable,
-					      MOVE_READ))
-			goto out;
-
-		if (mask->alphaMap &&
-		    !sna_drawable_move_to_cpu(mask->alphaMap->pDrawable,
-					      MOVE_READ))
-			goto out;
-	}
-
 	DBG(("%s: fallback -- fbComposite\n", __FUNCTION__));
-	fbComposite(op, src, mask, dst,
-		    src_x,  src_y,
-		    mask_x, mask_y,
-		    dst_x,  dst_y,
-		    width,  height);
+	sna_composite_fb(op, src, mask, dst, &region,
+			 src_x,  src_y,
+			 mask_x, mask_y,
+			 dst_x,  dst_y,
+			 width,  height);
 out:
 	REGION_UNINIT(NULL, &region);
 }
 
-static int16_t bound(int16_t a, uint16_t b)
-{
-	int v = (int)a + (int)b;
-	if (v > MAXSHORT)
-		return MAXSHORT;
-	return v;
-}
-
 static bool
 _pixman_region_init_clipped_rectangles(pixman_region16_t *region,
 				       unsigned int num_rects,
@@ -774,11 +892,11 @@ sna_composite_rectangles(CARD8		 op,
 		return;
 	}
 
-	DBG(("%s: drawable extents (%d, %d),(%d, %d) x %d\n",
+	DBG(("%s: drawable extents (%d, %d),(%d, %d) x %ld\n",
 	     __FUNCTION__,
 	     RegionExtents(&region)->x1, RegionExtents(&region)->y1,
 	     RegionExtents(&region)->x2, RegionExtents(&region)->y2,
-	     RegionNumRects(&region)));
+	     (long)RegionNumRects(&region)));
 
 	if (dst->pCompositeClip->data &&
 	    (!pixman_region_intersect(&region, &region, dst->pCompositeClip) ||
@@ -789,15 +907,15 @@ sna_composite_rectangles(CARD8		 op,
 		return;
 	}
 
-	DBG(("%s: clipped extents (%d, %d),(%d, %d) x %d\n",
+	DBG(("%s: clipped extents (%d, %d),(%d, %d) x %ld\n",
 	     __FUNCTION__,
 	     RegionExtents(&region)->x1, RegionExtents(&region)->y1,
 	     RegionExtents(&region)->x2, RegionExtents(&region)->y2,
-	     RegionNumRects(&region)));
+	     (long)RegionNumRects(&region)));
 
 	pixmap = get_drawable_pixmap(dst->pDrawable);
-	get_drawable_deltas(dst->pDrawable, pixmap, &dst_x, &dst_y);
-	pixman_region_translate(&region, dst_x, dst_y);
+	if (get_drawable_deltas(dst->pDrawable, pixmap, &dst_x, &dst_y))
+		pixman_region_translate(&region, dst_x, dst_y);
 
 	DBG(("%s: pixmap +(%d, %d) extents (%d, %d),(%d, %d)\n",
 	     __FUNCTION__, dst_x, dst_y,
@@ -823,32 +941,38 @@ sna_composite_rectangles(CARD8		 op,
 		goto fallback;
 	}
 
+	/* XXX xserver-1.8: CompositeRects is not tracked by Damage, so we must
+	 * manually append the damaged regions ourselves.
+	 */
+	DamageRegionAppend(&pixmap->drawable, &region);
 	boxes = pixman_region_rectangles(&region, &num_boxes);
 
 	/* If we going to be overwriting any CPU damage with a subsequent
 	 * operation, then we may as well delete it without moving it
 	 * first to the GPU.
 	 */
-	hint = PREFER_GPU;
+	hint = can_render(sna) ? PREFER_GPU : 0;
 	if (op <= PictOpSrc) {
 		if (priv->cpu_damage &&
 		    region_subsumes_damage(&region, priv->cpu_damage)) {
 			DBG(("%s: discarding existing CPU damage\n", __FUNCTION__));
 			if (priv->gpu_bo && priv->gpu_bo->proxy) {
+				assert(priv->gpu_damage == NULL);
 				kgem_bo_destroy(&sna->kgem, priv->gpu_bo);
 				priv->gpu_bo = NULL;
 			}
 			sna_damage_destroy(&priv->cpu_damage);
-			list_del(&priv->list);
+			list_del(&priv->flush_list);
 		}
-		if (region_subsumes_drawable(&region, &pixmap->drawable) ||
-		    box_inplace(pixmap, &region.extents)) {
+		if (region_subsumes_drawable(&region, &pixmap->drawable))
+			hint |= REPLACES;
+		if (hint & REPLACES || box_inplace(pixmap, &region.extents)) {
 			DBG(("%s: promoting to full GPU\n", __FUNCTION__));
 			if (priv->gpu_bo && priv->cpu_damage == NULL) {
+				assert(priv->gpu_bo->proxy == NULL);
 				sna_damage_all(&priv->gpu_damage,
 					       pixmap->drawable.width,
 					       pixmap->drawable.height);
-				priv->undamaged = false;
 			}
 		}
 		if (priv->cpu_damage == NULL) {
@@ -866,6 +990,8 @@ sna_composite_rectangles(CARD8		 op,
 		DBG(("%s: fallback due to no GPU bo\n", __FUNCTION__));
 		goto fallback;
 	}
+	if (hint & REPLACES)
+		kgem_bo_undo(&sna->kgem, bo);
 
 	if (!sna->render.fill_boxes(sna, op, dst->format, color,
 				    pixmap, bo, boxes, num_boxes)) {
@@ -886,7 +1012,6 @@ sna_composite_rectangles(CARD8		 op,
 				       pixmap->drawable.height);
 			sna_damage_destroy(damage == &priv->gpu_damage ?
 					   &priv->cpu_damage : &priv->gpu_damage);
-			priv->undamaged = false;
 		}
 
 		if (op <= PictOpSrc && bo == priv->gpu_bo) {
@@ -904,8 +1029,9 @@ sna_composite_rectangles(CARD8		 op,
 							     color->alpha,
 							     dst->format);
 			priv->clear = ok;
-			DBG(("%s: marking clear [%08x]? %d\n",
-			     __FUNCTION__, priv->clear_color, ok));
+			DBG(("%s: pixmap=%ld marking clear [%08x]? %d\n",
+			     __FUNCTION__, pixmap->drawable.serialNumber,
+			     priv->clear_color, ok));
 		}
 	}
 	goto done;
@@ -923,9 +1049,11 @@ fallback:
 	    !sna_drawable_move_to_cpu(dst->alphaMap->pDrawable, error))
 		goto done;
 
+	assert(pixmap->devPrivate.ptr);
+
 	if (op <= PictOpSrc) {
-		int nbox = REGION_NUM_RECTS(&region);
-		BoxPtr box = REGION_RECTS(&region);
+		int nbox = RegionNumRects(&region);
+		BoxPtr box = RegionRects(&region);
 		uint32_t pixel;
 
 		if (op == PictOpClear)
@@ -980,10 +1108,6 @@ fallback_composite:
 	}
 
 done:
-	/* XXX xserver-1.8: CompositeRects is not tracked by Damage, so we must
-	 * manually append the damaged regions ourselves.
-	 */
-	DamageRegionAppend(&pixmap->drawable, &region);
 	DamageRegionProcessPending(&pixmap->drawable);
 
 	pixman_region_fini(&region);
