@@ -17,24 +17,7 @@ static inline bool need_redirect(struct sna *sna, PixmapPtr dst)
 		dst->drawable.height > sna->render.max_3d_size);
 }
 
-static inline int vertex_space(struct sna *sna)
-{
-	return sna->render.vertex_size - sna->render.vertex_used;
-}
-static inline void vertex_emit(struct sna *sna, float v)
-{
-	assert(sna->render.vertex_used < sna->render.vertex_size);
-	sna->render.vertices[sna->render.vertex_used++] = v;
-}
-static inline void vertex_emit_2s(struct sna *sna, int16_t x, int16_t y)
-{
-	int16_t *v = (int16_t *)&sna->render.vertices[sna->render.vertex_used++];
-	assert(sna->render.vertex_used <= sna->render.vertex_size);
-	v[0] = x;
-	v[1] = y;
-}
-
-static inline float pack_2s(int16_t x, int16_t y)
+static force_inline float pack_2s(int16_t x, int16_t y)
 {
 	union {
 		struct sna_coordinate p;
@@ -45,19 +28,35 @@ static inline float pack_2s(int16_t x, int16_t y)
 	return u.f;
 }
 
-static inline int batch_space(struct sna *sna)
+static force_inline int vertex_space(struct sna *sna)
 {
+	return sna->render.vertex_size - sna->render.vertex_used;
+}
+static force_inline void vertex_emit(struct sna *sna, float v)
+{
+	assert(sna->render.vertex_used < sna->render.vertex_size);
+	sna->render.vertices[sna->render.vertex_used++] = v;
+}
+static force_inline void vertex_emit_2s(struct sna *sna, int16_t x, int16_t y)
+{
+	vertex_emit(sna, pack_2s(x, y));
+}
+
+static force_inline int batch_space(struct sna *sna)
+{
+	assert(sna->kgem.nbatch <= KGEM_BATCH_SIZE(&sna->kgem));
+	assert(sna->kgem.nbatch + KGEM_BATCH_RESERVED <= sna->kgem.surface);
 	return sna->kgem.surface - sna->kgem.nbatch - KGEM_BATCH_RESERVED;
 }
 
-static inline void batch_emit(struct sna *sna, uint32_t dword)
+static force_inline void batch_emit(struct sna *sna, uint32_t dword)
 {
 	assert(sna->kgem.mode != KGEM_NONE);
 	assert(sna->kgem.nbatch + KGEM_BATCH_RESERVED < sna->kgem.surface);
 	sna->kgem.batch[sna->kgem.nbatch++] = dword;
 }
 
-static inline void batch_emit_float(struct sna *sna, float f)
+static force_inline void batch_emit_float(struct sna *sna, float f)
 {
 	union {
 		uint32_t dw;
@@ -68,17 +67,18 @@ static inline void batch_emit_float(struct sna *sna, float f)
 }
 
 static inline bool
-is_gpu(DrawablePtr drawable)
+is_gpu(struct sna *sna, DrawablePtr drawable, unsigned prefer)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
 
-	if (priv == NULL || priv->clear)
+	if (priv == NULL || priv->clear || priv->cpu)
 		return false;
 
 	if (priv->cpu_damage == NULL)
 		return true;
 
-	if (priv->gpu_damage && !priv->gpu_bo->proxy)
+	if (priv->gpu_damage && !priv->gpu_bo->proxy &&
+	    (sna->render.prefer_gpu & prefer))
 		return true;
 
 	if (priv->cpu_bo && kgem_bo_is_busy(priv->cpu_bo))
@@ -105,15 +105,24 @@ static inline bool
 unattached(DrawablePtr drawable)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
-	return priv == NULL || (priv->gpu_damage == NULL && priv->cpu_damage);
+	return priv == NULL || (priv->gpu_damage == NULL && priv->cpu_damage && !priv->cpu_bo);
 }
 
 static inline bool
-picture_is_gpu(PicturePtr picture)
+picture_is_gpu(struct sna *sna, PicturePtr picture)
 {
 	if (!picture || !picture->pDrawable)
 		return false;
-	return is_gpu(picture->pDrawable);
+	return is_gpu(sna, picture->pDrawable, PREFER_GPU_RENDER);
+}
+
+static inline bool
+picture_is_cpu(struct sna *sna, PicturePtr picture)
+{
+	if (!picture->pDrawable)
+		return false;
+
+	return !is_gpu(sna, picture->pDrawable, PREFER_GPU_RENDER);
 }
 
 static inline bool sna_blt_compare_depth(DrawablePtr src, DrawablePtr dst)
@@ -139,6 +148,27 @@ sna_render_get_alpha_gradient(struct sna *sna)
 	return kgem_bo_reference(sna->render.alpha_cache.cache_bo);
 }
 
+static inline void
+sna_render_picture_extents(PicturePtr p, BoxRec *box)
+{
+	box->x1 = p->pDrawable->x;
+	box->y1 = p->pDrawable->y;
+	box->x2 = bound(box->x1, p->pDrawable->width);
+	box->y2 = bound(box->y1, p->pDrawable->height);
+
+	if (box->x1 < p->pCompositeClip->extents.x1)
+		box->x1 = p->pCompositeClip->extents.x1;
+	if (box->y1 < p->pCompositeClip->extents.y1)
+		box->y1 = p->pCompositeClip->extents.y1;
+
+	if (box->x2 > p->pCompositeClip->extents.x2)
+		box->x2 = p->pCompositeClip->extents.x2;
+	if (box->y2 > p->pCompositeClip->extents.y2)
+		box->y2 = p->pCompositeClip->extents.y2;
+
+	assert(box->x2 > box->x1 && box->y2 > box->y1);
+}
+
 static inline void
 sna_render_reduce_damage(struct sna_composite_op *op,
 			 int dst_x, int dst_y,
@@ -197,4 +227,44 @@ color_convert(uint32_t pixel,
 	return pixel;
 }
 
+inline static bool dst_use_gpu(PixmapPtr pixmap)
+{
+	struct sna_pixmap *priv = sna_pixmap(pixmap);
+	if (priv == NULL)
+		return false;
+
+	if (priv->cpu_bo && kgem_bo_is_busy(priv->cpu_bo))
+		return true;
+
+	if (priv->clear)
+		return false;
+
+	if (priv->gpu_bo && kgem_bo_is_busy(priv->gpu_bo))
+		return true;
+
+	return priv->gpu_damage && (!priv->cpu || !priv->cpu_damage);
+}
+
+inline static bool dst_use_cpu(PixmapPtr pixmap)
+{
+	struct sna_pixmap *priv = sna_pixmap(pixmap);
+	if (priv == NULL || priv->shm)
+		return true;
+
+	return priv->cpu_damage && priv->cpu;
+}
+
+inline static bool dst_is_cpu(PixmapPtr pixmap)
+{
+	struct sna_pixmap *priv = sna_pixmap(pixmap);
+	return priv == NULL || DAMAGE_IS_ALL(priv->cpu_damage);
+}
+
+inline static bool
+untransformed(PicturePtr p)
+{
+	return !p->transform || pixman_transform_is_int_translate(p->transform);
+}
+
+
 #endif /* SNA_RENDER_INLINE_H */
